/*
Copyright (c) 2009-2010, Stephane Sudre
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
- Neither the name of the WhiteBox nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#import "ICSourceTextViewDelegate.h"

#import "NSRangeUtilities.h"

#import "NSString+Ranges.h"

#import "ICSourceTextView+Constants.h"

NSString * IC_SOURCE_TEXTVIEW_FUNCTIONS_LIST_DID_CHANGE=@"IC_SOURCE_TEXTVIEW_FUNCTIONS_LIST_DID_CHANGE";

NSString * const FUNCTION_NAME_KEY=@"FUNCTION_NAME_KEY";

NSString * const FUNCTION_PARAMETERS_KEY=@"FUNCTION_PARAMETERS_KEY";

NSString * const FUNCTION_PROTOTYPE_KEY=@"FUNCTION_PROTOTYPE_KEY";

NSString * const FUNCTION_BLOCK_KEY=@"FUNCTION_BLOCK_KEY";


NSString * const IC_SOURCETEXTVIEW_DELEGATE_EDITOR_FONT=@"javascript.editor.font";

#define IC_SOURCETEXTVIEW_DELEGATE_DELAYED_POPUP_UPDATE	0.1f

@implementation ICSourceTextViewDelegate

- (id) init
{
	self=[super init];
	
	if (self!=nil)
	{
		nonSeparatorsSet_=[[NSCharacterSet characterSetWithCharactersInString:@"_ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890"] retain];


		keywords_=[[NSArray alloc] initWithObjects: @"if",
													@"else",
													@"for",
													@"in",
													@"while",
													@"do",
													@"continue",
													@"break",
													@"with",
													@"try",
													@"catch",
													@"switch",
													@"case",
													@"new",
													@"var",
													@"function",
													@"return",
													@"this",
													@"delete",
													@"true",
													@"false",
													@"void",
													@"throw",
													@"typeof",
													@"const",
													@"default",
													nil];
		
		
		distributionKeywords_=[[NSArray alloc] initWithObjects: @"system",
																@"compareVersions",
																@"numericalCompare",
																@"gestalt",
																@"localizedStandardString",
																@"localizedStandardStringWithFormat",
																@"localizedString",
																@"localizedStringWithFormat",
																@"log",
																@"propertiesOf",
																@"run",
																@"runOnce",
																@"sysctl",
																@"all",
																@"fromIdentifier",
																@"fromPID",
																@"fileExistsAtPath",
																@"plistAtPath",
																@"bundleAtPath",
																@"fromPath",
																@"matchingClass",
																@"matchingName",
																@"childrenOf",
																@"parentsOf",
																@"applications",
																@"defaults",
																@"files",
																@"ioregistry",
																@"users.desktopSessionCount",
																@"version",
																@"title",
																@"description",
																@"availableKilobytes",
																@"systemVersion",
																@"ProductName",
																@"ProductVersion",
																@"BuildVersion",
																@"my",
																@"choice",
																@"tooltip",
																@"selected",
																@"packageUpgradeAction",
																@"result",
																@"message",
																@"type",
																@"target",
																@"mountpoint",
																@"isServer",
																@"receiptForIdentifier",
																@"packages",
																@"ignoreContents",
																
																nil];
		
		syntaxAttributeDictionary_=[[NSMutableDictionary alloc] initWithCapacity:10];
		
		if (syntaxAttributeDictionary_!=nil)
		{
			[syntaxAttributeDictionary_ setObject:[NSDictionary dictionaryWithObjectsAndKeys:[NSColor colorWithCalibratedRed:0.0f/255.0f green:116.0f/255.0f blue:0.0f/255.0f alpha:1.0f],NSForegroundColorAttributeName,nil] forKey:@"COMMENTS"];
			
			[syntaxAttributeDictionary_ setObject:[NSDictionary dictionaryWithObjectsAndKeys:[NSColor colorWithCalibratedRed:196.0f/255.0f green:26.0f/255.0f blue:22.0f/255.0f alpha:1.0f],NSForegroundColorAttributeName,nil] forKey:@"STRINGS"];
			
			[syntaxAttributeDictionary_ setObject:[NSDictionary dictionaryWithObjectsAndKeys:[NSColor colorWithCalibratedRed:170.0f/255.0f green:13.0f/255.0f blue:145.0f/255.0f alpha:1.0f],NSForegroundColorAttributeName,nil] forKey:@"KEYWORDS"];
			
			[syntaxAttributeDictionary_ setObject:[NSDictionary dictionaryWithObjectsAndKeys:[NSColor colorWithCalibratedRed:46.0f/255.0f green:13.0f/255.0f blue:110.0f/255.0f alpha:1.0f],NSForegroundColorAttributeName,nil] forKey:@"DISTRIBUTION"];
		}
	}
	
	return self;
}

- (void) dealloc
{
	
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	
	[super dealloc];
}

#pragma mark -

- (void) awakeFromNib
{
	NSUserDefaults * tUserDefaults;
	NSString * tString;
	NSString * tFontName=@"Monaco";
	CGFloat tFontSize=10.0f;
	
	tUserDefaults=[NSUserDefaults standardUserDefaults];
	
	// Text Font
	
	tString=[tUserDefaults stringForKey:IC_SOURCETEXTVIEW_DELEGATE_EDITOR_FONT];
	
	if (tString!=nil)
	{
		NSArray * tArray;
		
		tArray=[tString componentsSeparatedByString:@" - "];
		
		if (tArray!=nil && [tArray count]==2)
		{
			tFontName=[tArray objectAtIndex:0];
			
			if ([tFontName length]<3)
			{
				tFontName=@"Monaco";
			}
			
			tFontSize=[[tArray objectAtIndex:1] floatValue];
			
			if (tFontSize<2.0f || tFontSize>200.0f)
			{
				tFontSize=10.0f;
			}
		}
	}
	
	[textView_ setFont:[NSFont fontWithName:tFontName size:tFontSize]];
	
	[functionsPopupButton_ removeAllItems];
	
	// Register for Notifications
		
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(textViewDidChangeFont:)
												 name:ICSOURCETEXTVIEW_NOTIFICATION_FONT_DID_CHANGE
											   object:textView_];
	
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(textViewDidChangeSelection:)
												 name:NSTextViewDidChangeSelectionNotification
											   object:textView_];
	
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(textDidChange:)
												 name:NSTextDidChangeNotification
											   object:textView_];
}

#pragma mark -

+ (NSMutableDictionary *) functionDictionaryForString:(NSString *) inString
{
	NSMutableDictionary * tFunctionDictionary=nil;
	
	if (inString!=nil)
	{
		NSUInteger tLength;
		NSRange tFunctionRange;
		
		tLength=[inString length];
		
		tFunctionRange=[inString rangeOfString:@"function" options:0 range:NSMakeRange(0,tLength)];
		
		if (tFunctionRange.location!=NSNotFound)
		{
			NSUInteger tWhiteSpaceIndex;
			BOOL notAFunction=NO;
			
			// Check that there is only white space before the keyword
			
			tWhiteSpaceIndex=tFunctionRange.location;
			
			while (tWhiteSpaceIndex>1)
			{
				unichar tChar;

				tChar=[inString characterAtIndex:tWhiteSpaceIndex-1];
				
				if (tChar==' ' ||
					tChar=='\t')
				{
					tWhiteSpaceIndex--;
				}
				else
				{
					notAFunction=YES;
					
					break;
				}
			}
			
			if (notAFunction==NO)
			{
				// Find the function name
				
				NSUInteger tFunctionNameIndex;
				
				tFunctionNameIndex=NSMaxRange(tFunctionRange);
				
				// At least one whitespace before the function name
				
				if (tFunctionNameIndex<tLength)
				{
					unichar tChar;
					
					tChar=[inString characterAtIndex:tFunctionNameIndex];
				
					if (tChar==' ' || tChar=='\t' || tChar=='\r' || tChar=='\n')
					{
						notAFunction=YES;
						
						// Find the first (
						
						tFunctionNameIndex++;
						
						while (tFunctionNameIndex<tLength)
						{
							tChar=[inString characterAtIndex:tFunctionNameIndex];
							
							if (tChar==' ' ||
								tChar=='\t' || 
								tChar=='\n' ||
								tChar=='\r' ||
								tChar=='_' ||
								(tChar>='A' && tChar<='Z') ||
								(tChar>='a' && tChar<='z') ||
								(tChar>='0' && tChar<='9'))
							{
								tFunctionNameIndex++;
							}
							else if (tChar=='(')
							{
								notAFunction=NO;
								
								tFunctionNameIndex--;
								
								break;
							}
							else
							{
								break;
							}
						}
						
						if (notAFunction==NO && tFunctionNameIndex<(tLength-2))
						{
							NSMutableString * tFunctionName;
					
							tFunctionName=[[[inString substringWithRange:NSMakeRange(NSMaxRange(tFunctionRange),tFunctionNameIndex-NSMaxRange(tFunctionRange)+1)] mutableCopy] autorelease];
							
							if (tFunctionName!=nil)
							{
								NSUInteger tNameLength;
								
								CFStringTrimWhitespace((CFMutableStringRef) tFunctionName);
								
								tNameLength=[tFunctionName length];
								
								if (tNameLength>0)
								{
									tChar=[tFunctionName characterAtIndex:0];
									
									if (tChar<'0' || tChar>'9')
									{
										// Check that we don't have a '\r' or '\n' or '\t' or ' ' in the name
									
										if ([tFunctionName rangeOfCharacterFromSet:[NSCharacterSet characterSetWithCharactersInString:@" \t\r\n"]].location==NSNotFound)
										{
											NSMutableString * tParametersString;
											
											tParametersString=[[[inString substringWithRange:NSMakeRange(tFunctionNameIndex+1,tLength-tFunctionNameIndex-1)] mutableCopy] autorelease];
											
											if (tParametersString!=nil)
											{
												NSUInteger tParametersLength;
												
												CFStringTrimWhitespace((CFMutableStringRef) tParametersString);
												
												tParametersLength=[tParametersString length];
												
												if (tParametersLength>=2)
												{
													tChar=[tParametersString characterAtIndex:tParametersLength-1];
													
													if (tChar==')')
													{
														NSMutableArray * tParameters;
														
														tParameters=[NSMutableArray array];
														
														if ((tParametersLength-2)>0)
														{
															NSString * tSubString;
														
															tSubString=[tParametersString substringWithRange:NSMakeRange(1,tParametersLength-2)];
														
															if ([tSubString length]>0)
															{
																NSArray * tParametersArray;
																
																tParametersArray=[tSubString componentsSeparatedByString:@","];
																
																if (tParametersArray!=nil)
																{
																	NSUInteger tParametersCount,tParametersIndex;
																	
																	tParametersCount=[tParametersArray count];
																	
																	for(tParametersIndex=0;tParametersIndex<tParametersCount;tParametersIndex++)
																	{
																		NSMutableString * tSingleParameterString;
																		
																		tSingleParameterString=[[[tParametersArray objectAtIndex:tParametersIndex] mutableCopy] autorelease];
																		
																		if (tSingleParameterString!=nil)
																		{
																			NSUInteger tSingleParameterLength;
																			
																			CFStringTrimWhitespace((CFMutableStringRef) tSingleParameterString);
																			
																			tSingleParameterLength=[tSingleParameterString length];
																			
																			if (tSingleParameterLength==0)
																			{
																				if (tParametersCount!=1)
																				{
																					return nil;
																				}
																			}
																			else
																			{
																				// Check that a parameter name only contains valid characters
																				
																				NSUInteger tCheckIndex;
																				
																				for(tCheckIndex=0;tCheckIndex<tSingleParameterLength;tCheckIndex++)
																				{
																					tChar=[tSingleParameterString characterAtIndex:tCheckIndex];
																					
																					if ((tChar>='A' && tChar<='Z') ||
																						(tChar>='a' && tChar<='z') ||
																						 tChar=='_')
																					{
																					}
																					else if (tChar>='0' && tChar<='9')
																					{
																						if (tCheckIndex==0)
																						{
																							return nil;
																						}
																					}
																					else
																					{
																						return nil;
																					}
																				}
																				
																				[tParameters addObject:tSingleParameterString];
																			}
																		}
																	}
																}
															}
														}
														
														if (tParameters!=nil)
														{
															tFunctionDictionary=[NSMutableDictionary dictionaryWithObjectsAndKeys:tFunctionName,FUNCTION_NAME_KEY,
																																  tParameters,FUNCTION_PARAMETERS_KEY,
																																  nil];
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
		else
		{
			// Not a function
		}
	}

	return tFunctionDictionary;
}

- (void) delayedParseFunctions:(id) anObject
{
	NSUInteger tRangeCount;
	
	tRangeCount=[searchableRangesArray_ count];
	
	[functions_ release];
	
	functions_=[[NSMutableArray alloc] initWithCapacity:10];
	
	if (tRangeCount>0)
	{
		NSMutableArray * tRootBlocksArray;
		NSString * tSourceCode;
		NSUInteger tLength;
		NSUInteger tRangeIndex=0;
		NSValue * tValue;
		NSRange tSearchRange;
		NSUInteger tRootBlocksCount;
		
		tSourceCode=[textView_ string];
	
		tLength=[tSourceCode length];
	
		tRootBlocksArray=[NSMutableArray array];
		
		tValue=[searchableRangesArray_ objectAtIndex:0];
			
		tSearchRange=[tValue rangeValue];
		
		do
		{
			NSRange tFoundRange;
			
			tFoundRange=[tSourceCode rangeOfString:@"{" options:0 range:tSearchRange];
			
			if (tFoundRange.location==NSNotFound)
			{
				tRangeIndex++;
				
				if (tRangeIndex<tRangeCount)
				{
					tValue=[searchableRangesArray_ objectAtIndex:tRangeIndex];
			
					tSearchRange=[tValue rangeValue];
				}
				else
				{
					break;
				}
			}
			else
			{
				NSUInteger tMatchLocation;
				NSRange tBlockRange;
				NSValue * tBlockValue;
				
				tBlockRange.location=tFoundRange.location;
				
				tMatchLocation=[self findForwardMatchingCharacter:'}' inRanges:searchableRangesArray_ startingAt:tBlockRange.location];
				
				if (tMatchLocation==NSNotFound)
				{
					tBlockRange.length=tLength-tBlockRange.location;
					
					tMatchLocation=tLength-1;
					
					tBlockValue=[NSValue valueWithRange:tBlockRange];
					
					if (tBlockValue!=nil)
					{
						[tRootBlocksArray addObject:tBlockValue];
					}
					
					break;
				}
				else
				{
					tBlockRange.length=tMatchLocation-tBlockRange.location+1;
					
					tBlockValue=[NSValue valueWithRange:tBlockRange];
					
					if (tBlockValue!=nil)
					{
						[tRootBlocksArray addObject:tBlockValue];
					}
					
					if ((tMatchLocation+1)<NSMaxRange(tSearchRange))
					{
						tSearchRange.length=NSMaxRange(tSearchRange)-tMatchLocation-1;
					
						tSearchRange.location=tMatchLocation+1;
					}
					else
					{
						tRangeIndex++;
				
						if (tRangeIndex<tRangeCount)
						{
							tValue=[searchableRangesArray_ objectAtIndex:tRangeIndex];
					
							tSearchRange=[tValue rangeValue];
						}
						else
						{
							break;
						}
					}
				}
			}
		}
		while (1);
		
		tRootBlocksCount=[tRootBlocksArray count];
		
		if (tRootBlocksCount>0)
		{
			NSUInteger tRootBlockIndex=0;
			NSRange tPreviousBlockRange=NSMakeRange(0,0);
			
			while (tRootBlockIndex<tRootBlocksCount)
			{
				NSRange tBlockRange;
				
				tValue=[tRootBlocksArray objectAtIndex:tRootBlockIndex];
			
				tBlockRange=[tValue rangeValue];
				
				if (tBlockRange.location>=(tPreviousBlockRange.location+12))	// Minimum distance required for function X()
				{
					NSRange tFoundRange;
					
					tSearchRange=NSMakeRange(tPreviousBlockRange.location,tBlockRange.location-tPreviousBlockRange.location);
					
					tFoundRange=[tSourceCode rangeOfString:@"function" options:NSBackwardsSearch range:tSearchRange];
					
					if (tFoundRange.location!=NSNotFound)
					{
						// Find the beginning of the line
						
						NSUInteger tCharIndex;
						NSRange tRestrictedRange;
						
						tCharIndex=tFoundRange.location;
						
						while (tCharIndex>0)
						{
							unichar tChar;
							
							tChar=[tSourceCode characterAtIndex:tCharIndex-1];
							
							if (tChar=='\n' || tChar=='\r')
							{
								break;
							}
							
							tCharIndex--;
						}
						
						tRestrictedRange.location=tCharIndex;
						
						tRestrictedRange.length=tBlockRange.location-tCharIndex;
						
						// Check this does not intersect a string range?
						
						if ([NSRangeUtilities range:tRestrictedRange intersectsSortedRanges:stringsRangesArray_]==NO)
						{
							NSArray * tIntersectedCommentsRange;
							NSString * tExcerptString;
							
							tIntersectedCommentsRange=[NSRangeUtilities sortedRanges:commentsRangesArray_ intersectingRange:tRestrictedRange];
							
							// Take the excerpt and remove the comments
							
							tExcerptString=[tSourceCode substringWithRange:tRestrictedRange excludingSortedRanges:tIntersectedCommentsRange];
							
							if (tExcerptString!=nil)
							{
								NSMutableDictionary * tFunctionDictionary;
								
								tFunctionDictionary=[ICSourceTextViewDelegate functionDictionaryForString:tExcerptString];
								
								if (tFunctionDictionary!=nil)
								{
									[tFunctionDictionary setObject:[NSValue valueWithRange:tRestrictedRange] forKey:FUNCTION_PROTOTYPE_KEY];
									
									[tFunctionDictionary setObject:[NSValue valueWithRange:NSUnionRange(tRestrictedRange,tBlockRange)] forKey:FUNCTION_BLOCK_KEY];
									
									[functions_ addObject:tFunctionDictionary];
								}
							}
						}
					}
				}
				
				tPreviousBlockRange=tBlockRange;
				
				tRootBlockIndex++;
			}
		}
	}
	
	// Update the functions menu
	
	[functionsPopupButton_ removeAllItems];
	
	if (functions_!=nil)
	{
		NSEnumerator * tEnumerator;
		
		
		tEnumerator=[functions_ objectEnumerator];
		
		if (tEnumerator!=nil)
		{
			NSDictionary * tDictionary;
			
			while (tDictionary=[tEnumerator nextObject])
			{
				NSString * tTitle;
				
				tTitle=[tDictionary objectForKey:FUNCTION_NAME_KEY];
				
				if (tTitle!=nil)
				{
					[functionsPopupButton_ addItemWithTitle:tTitle];
				}
			}
		}
	}

	[NSObject cancelPreviousPerformRequestsWithTarget:self selector:@selector(updateFunctionsMenuSelectedItem:) object:nil];
			
	[self updateFunctionsMenuSelectedItem:nil];
	
	[[NSNotificationCenter defaultCenter] postNotificationName:IC_SOURCE_TEXTVIEW_FUNCTIONS_LIST_DID_CHANGE object:self];
}

- (void) updateFunctionsMenuSelectedItem:(id) anObject
{
	NSRect tFrame;
	NSArray * tSelectedRanges; 
	NSUInteger tSelectedFunctionIndex=NSNotFound;
	
	tSelectedRanges=[textView_ selectedRanges];
	
	if (tSelectedRanges!=nil && [tSelectedRanges count]>0)
	{
		NSValue * tValue;
		NSRange tRange;
		NSUInteger tLocation;
		
		
		tValue=[tSelectedRanges objectAtIndex:0];
		
		tRange=[tValue rangeValue];
		
		tLocation=tRange.location;
		
		tSelectedFunctionIndex=[NSRangeUtilities indexOfRangeIncludingLocation:tLocation withinRanges:[self functionsRanges]];
	}
	
	if (tSelectedFunctionIndex!=NSNotFound)
	{
		[functionsPopupButton_ selectItemAtIndex:tSelectedFunctionIndex];
	}
	else
	{
		[functionsPopupButton_ selectItemAtIndex:-1];
	}
	
	tFrame=[functionsPopupButton_ frame];
			
	[functionsPopupButton_ sizeToFit];
	
	tFrame.size.width=NSWidth([functionsPopupButton_ frame])+5.0f;
	
	[functionsPopupButton_ setFrame:tFrame];

	[[functionsPopupButton_ superview] setNeedsDisplay:YES];
}

- (void) parseSourceCode
{
	NSString * tSourceCode;
	NSUInteger tLength;
	
	tSourceCode=[textView_ string];
	
	tLength=[tSourceCode length];
	
	[commentsRangesArray_ autorelease];
	
	commentsRangesArray_=[[NSMutableArray alloc] initWithCapacity:10];
	
	[stringsRangesArray_ autorelease];
	
	stringsRangesArray_=[[NSMutableArray alloc] initWithCapacity:10];
	
	[searchableRangesArray_ release];
	
	searchableRangesArray_=nil;
	
	if (tLength>0 && commentsRangesArray_!=nil && stringsRangesArray_!=nil)
	{
		NSUInteger tIndex=0;
		NSValue * tValue;
	
		while (tIndex<tLength)
		{
			unichar tChar;
			
			tChar=[tSourceCode characterAtIndex:tIndex];
			
			if (tChar=='/')
			{
				tIndex++;
				
				if (tIndex<tLength)
				{
					tChar=[tSourceCode characterAtIndex:tIndex];
					
					if (tChar=='*')
					{
						NSRange tMultilineCommentRange;
					
						tMultilineCommentRange.location=tIndex-1;
						tMultilineCommentRange.length=0;
						
						// Look for the first closing */
						
						tIndex++;
						
						if (tIndex<tLength)
						{
							while (tIndex<tLength)
							{
								tChar=[tSourceCode characterAtIndex:tIndex];
								
								if (tChar=='*')
								{
									tIndex++;
									
									if (tIndex<tLength)
									{
										tChar=[tSourceCode characterAtIndex:tIndex];
								
										if (tChar=='/')
										{
											tMultilineCommentRange.length=tIndex-tMultilineCommentRange.location+1;
											
											tValue=[NSValue valueWithRange:tMultilineCommentRange];
											
											if (tValue!=nil)
											{
												[commentsRangesArray_ addObject:tValue];
												
												tIndex++;
											}
											else
											{
												// Log Error
											
												// A COMPLETER
											}
											
											break;
										}
										else
										{
											if (tChar!='*')
											{
												tIndex++;
											}
										}
									}
									else
									{
										break;
									}
								}
								else
								{
									tIndex++;
								}
							}
						}
						
						if (tIndex>=tLength && tMultilineCommentRange.length==0)
						{
							tMultilineCommentRange.length=tLength-tMultilineCommentRange.location;
											
							tValue=[NSValue valueWithRange:tMultilineCommentRange];
							
							if (tValue!=nil)
							{
								[commentsRangesArray_ addObject:tValue];
							}
							else
							{
								// Log Error
								
								// A COMPLETER
							}
							
							break;
						}
					}
					else if (tChar=='/')
					{
						NSRange tSinglelineCommentRange;
						
						tSinglelineCommentRange.location=tIndex-1;
						tSinglelineCommentRange.length=0;
						
						// We need to look for the first end of line
						
						tIndex++;
						
						if (tIndex<tLength)
						{
							while (tIndex<tLength)
							{
								tChar=[tSourceCode characterAtIndex:tIndex];
								
								if (tChar=='\n' || tChar=='\r')
								{
									tSinglelineCommentRange.length=tIndex-tSinglelineCommentRange.location;
											
									tValue=[NSValue valueWithRange:tSinglelineCommentRange];
									
									if (tValue!=nil)
									{
										[commentsRangesArray_ addObject:tValue];
									}
									else
									{
										// Log Error
										
										// A COMPLETER
									}
									
									tIndex++;
									
									break;
								}
								else
								{
									tIndex++;
								}
							}
						}
						
						if (tIndex>=tLength && tSinglelineCommentRange.length==0)
						{
							tSinglelineCommentRange.length=tLength-tSinglelineCommentRange.location;
											
							tValue=[NSValue valueWithRange:tSinglelineCommentRange];
							
							if (tValue!=nil)
							{
								[commentsRangesArray_ addObject:tValue];
							}
							else
							{
								// Log Error
								
								// A COMPLETER
							}
							
							break;
						}
					}
				}
			}
			else if (tChar=='\"')
			{
				NSRange tStringRange;
						
				tStringRange.location=tIndex;
				tStringRange.length=0;
				
				// We need to look for the first end of line
				
				tIndex++;
				
				if (tIndex<tLength)
				{
					// Find the first "
					
					while (tIndex<tLength)
					{
						tChar=[tSourceCode characterAtIndex:tIndex];
						
						if (tChar=='\\')
						{
							tIndex+=2;
						}
						else if (tChar=='\"')
						{
							tChar=[tSourceCode characterAtIndex:tIndex-1];
							
							tStringRange.length=tIndex-tStringRange.location+1;
									
							tValue=[NSValue valueWithRange:tStringRange];
							
							if (tValue!=nil)
							{
								[stringsRangesArray_ addObject:tValue];
							}
							else
							{
								// Log Error
								
								// A COMPLETER
							}
							
							tIndex++;
							
							break;
						}
						else
						{
							tIndex++;
						}
					}
				}
				
				if (tIndex>=tLength && tStringRange.length==0)
				{
					tStringRange.length=tLength-tStringRange.location;
									
					tValue=[NSValue valueWithRange:tStringRange];
					
					if (tValue!=nil)
					{
						[stringsRangesArray_ addObject:tValue];
					}
					else
					{
						// Log Error
						
						// A COMPLETER
					}
					
					break;
				}
			}
			else if (tChar=='\'')
			{
				NSRange tStringRange;
						
				tStringRange.location=tIndex;
				tStringRange.length=0;
				
				// We need to look for the first end of line
				
				tIndex++;
				
				if (tIndex<tLength)
				{
					// Find the first '
					
					while (tIndex<tLength)
					{
						tChar=[tSourceCode characterAtIndex:tIndex];
						
						if (tChar=='\\')
						{
							tIndex+=2;
						}
						else if (tChar=='\'')
						{
							tChar=[tSourceCode characterAtIndex:tIndex-1];
							
							tStringRange.length=tIndex-tStringRange.location+1;
									
							tValue=[NSValue valueWithRange:tStringRange];
							
							if (tValue!=nil)
							{
								[stringsRangesArray_ addObject:tValue];
							}
							else
							{
								// Log Error
								
								// A COMPLETER
							}
							
							tIndex++;
							
							break;
						}
						else
						{
							tIndex++;
						}
					}
				}
				
				if (tIndex>=tLength && tStringRange.length==0)
				{
					tStringRange.length=tLength-tStringRange.location;
									
					tValue=[NSValue valueWithRange:tStringRange];
					
					if (tValue!=nil)
					{
						[stringsRangesArray_ addObject:tValue];
					}
					else
					{
						// Log Error
						
						// A COMPLETER
					}
					
					break;
				}
			}
			else
			{
				tIndex++;
			}
		}
	}
	
	// Colorize code
	
	NSTextStorage * tTextStorage;
	NSLayoutManager * tLayoutManager;
	
	tTextStorage=[textView_ textStorage];
	
	tLayoutManager  = [[tTextStorage layoutManagers] objectAtIndex: 0];
	
	// Remove all attributes
	
	[tLayoutManager removeTemporaryAttribute:NSForegroundColorAttributeName forCharacterRange:NSMakeRange(0,tLength)]; 
	
	if (tLength>0)
	{
		NSValue * tValue;
		NSRange tRange;
		NSDictionary * tAttributeDictionary;
		
		// Comments
		
		if (commentsRangesArray_!=nil)
		{
			NSEnumerator * tEnumerator;
			
			tEnumerator=[commentsRangesArray_ objectEnumerator];
			
			if (tEnumerator!=nil)
			{
				tAttributeDictionary=[syntaxAttributeDictionary_ objectForKey:@"COMMENTS"];
				
				while (tValue=[tEnumerator nextObject])
				{
					tRange=[tValue rangeValue];
					
					[tLayoutManager addTemporaryAttributes: tAttributeDictionary forCharacterRange: tRange];
				}
			}
		}	
	
		// Strings
		
		if (stringsRangesArray_!=nil)
		{
			NSEnumerator * tEnumerator;
			
			tEnumerator=[stringsRangesArray_ objectEnumerator];
			
			if (tEnumerator!=nil)
			{
				tAttributeDictionary=[syntaxAttributeDictionary_ objectForKey:@"STRINGS"];
				
				while (tValue=[tEnumerator nextObject])
				{
					tRange=[tValue rangeValue];
					
					// A COMPLETER
					
					[tLayoutManager addTemporaryAttributes: tAttributeDictionary forCharacterRange: tRange];
				}
			}
		}
		
		NSMutableArray * tExcludedRangesArray;
		
		tExcludedRangesArray=[NSMutableArray arrayWithArray:commentsRangesArray_];
		
		[tExcludedRangesArray addObjectsFromArray:stringsRangesArray_];
		
		
		
		searchableRangesArray_=[[NSRangeUtilities rangesFromRange:NSMakeRange(0,tLength) excludingRanges:tExcludedRangesArray] retain];
		
		if (searchableRangesArray_!=nil && [searchableRangesArray_ count]>0)
		{
			NSEnumerator * tEnumerator;
			
			tEnumerator=[searchableRangesArray_ objectEnumerator];
			
			if (tEnumerator!=nil)
			{
				while (tValue=[tEnumerator nextObject])
				{
					tRange=[tValue rangeValue];
					
					if (keywords_!=nil && [keywords_ count]>0)
					{
						tAttributeDictionary=[syntaxAttributeDictionary_ objectForKey:@"KEYWORDS"];
						
						[self highlightKeywords:keywords_ inRange:tRange withAttributes:tAttributeDictionary];
					}
					
					if (distributionKeywords_!=nil && [distributionKeywords_ count]>0)
					{
						tAttributeDictionary=[syntaxAttributeDictionary_ objectForKey:@"DISTRIBUTION"];
						
						[self highlightKeywords:distributionKeywords_ inRange:tRange withAttributes:tAttributeDictionary];
					}
				}
			}
		}
	}
}

- (void) highlightKeywords:(NSArray *) inKeywordsArray inRange:(NSRange) inRange withAttributes:(NSDictionary *) inAttributes
{
	if (inKeywordsArray!=nil && inRange.length>0)
	{
		NSEnumerator * tEnumerator;
		
		tEnumerator=[inKeywordsArray objectEnumerator];
			
		if (tEnumerator!=nil)
		{
			NSTextStorage * tTextStorage;
			NSLayoutManager * tLayoutManager;
			NSString * tSourceCode;
			NSString * tKeyword;
			
			tTextStorage=[textView_ textStorage];
	
			tLayoutManager = [[tTextStorage layoutManagers] objectAtIndex: 0];
			
			tSourceCode=[textView_ string];
			
			while (tKeyword=[tEnumerator nextObject])
			{
				NSRange tSearchableRange;
				
				tSearchableRange=inRange;
				
				while (tSearchableRange.location<NSMaxRange(inRange))
				{
					NSRange tFoundRange;
					
					tFoundRange=[tSourceCode rangeOfString:tKeyword options:0 range:tSearchableRange];
				
					if (tFoundRange.location!=NSNotFound)
					{
						BOOL tProblem=NO;
						
						if (tFoundRange.location>inRange.location)
						{
							tProblem=[nonSeparatorsSet_ characterIsMember:[tSourceCode characterAtIndex:(tFoundRange.location-1)]];
						}
						
						if (tProblem==NO)
						{
							if (NSMaxRange(tFoundRange)<NSMaxRange(inRange))
							{
								tProblem=[nonSeparatorsSet_ characterIsMember:[tSourceCode characterAtIndex:NSMaxRange(tFoundRange)]];
							}
						}
						
						if (tProblem==NO)
						{
							[tLayoutManager addTemporaryAttributes:inAttributes forCharacterRange:tFoundRange];
						}
						
						tSearchableRange.length=NSMaxRange(tSearchableRange)-NSMaxRange(tFoundRange);
						
						tSearchableRange.location=NSMaxRange(tFoundRange);
					}
					else
					{
						break;
					}
				}
			}
		}
	}
}

- (NSUInteger) findForwardMatchingCharacter:(unichar) inCharacter inRanges:(NSArray *) inRangesArray startingAt:(NSUInteger) inStartingIndex
{
	NSUInteger tRangeIndex;
	
	tRangeIndex=[NSRangeUtilities indexOfRangeIncludingLocation:inStartingIndex withinRanges:inRangesArray];
	
	if (tRangeIndex!=NSNotFound)
	{
		NSValue * tValue;
		NSInteger tRangesCount;
		NSCharacterSet * tLookingForSet;
		NSString * tSourceCode;
		NSUInteger tBracesDepth=0;
		NSUInteger tParenthesisDepth=0;
		
		if (inCharacter==')')
		{
			tParenthesisDepth=1;
		}
		else if (inCharacter=='}')
		{
			tBracesDepth=1;
		}
		else
		{
			return NSNotFound;
		}
		
		tSourceCode=[textView_ string];
		
		tLookingForSet=[NSCharacterSet characterSetWithCharactersInString:@"(){}"];
		
		tRangesCount=[inRangesArray count];
		
		while (tRangeIndex<tRangesCount)
		{
			NSRange tPartialLookRange;
			BOOL tContinue=YES;
			
			tValue=[inRangesArray objectAtIndex:tRangeIndex];
			
			tPartialLookRange=[tValue rangeValue];
			
			if (tPartialLookRange.location<=inStartingIndex)
			{
				if (inStartingIndex==(NSMaxRange(tPartialLookRange)-1))
				{
					tRangeIndex++;
					
					continue;
				}
				
				tPartialLookRange.length=NSMaxRange(tPartialLookRange)-inStartingIndex-1;
				
				tPartialLookRange.location=inStartingIndex+1;
			}
			
			while (tContinue==YES)
			{
				NSRange tFoundRange;
				
				tFoundRange=[tSourceCode rangeOfCharacterFromSet:tLookingForSet options:0 range:tPartialLookRange];
				
				if (tFoundRange.location==NSNotFound)
				{
					break;
				}
				else
				{
					unichar tFoundCharacter;
					
					tFoundCharacter=[tSourceCode characterAtIndex:tFoundRange.location];
					
					switch (tFoundCharacter)
					{
						case '{':
						
							if (tParenthesisDepth!=0)
							{
								return NSNotFound;
							}
						
							tBracesDepth++;
							
							break;
						
						case '}':
						
							if (tParenthesisDepth!=0)
							{
								return NSNotFound;
							}
						
							tBracesDepth--;
							
							if (tBracesDepth==0 && inCharacter=='}')
							{
								return tFoundRange.location;
							}
							
							break;
							
						case '(':
							
							tParenthesisDepth++;
							
							break;
							
						case ')':
							
							tParenthesisDepth--;
							
							if (tParenthesisDepth==0 && inCharacter==')')
							{
								return tFoundRange.location;
							}
							
							break;
					}
					
					if ((tFoundRange.location+1)>=NSMaxRange(tPartialLookRange))
					{
						tContinue=NO;
					}
					else
					{
						tPartialLookRange.length=NSMaxRange(tPartialLookRange)-tFoundRange.location-1;
						
						tPartialLookRange.location=tFoundRange.location+1;
					}
				}
			}
			
			tRangeIndex++;
		}
	}
	
	return NSNotFound;
}

- (NSUInteger) findBackwardMatchingCharacter:(unichar) inCharacter inRanges:(NSArray *) inRangesArray startingAt:(NSUInteger) inStartingIndex
{
	NSUInteger tRangeIndex;
	
	tRangeIndex=[NSRangeUtilities indexOfRangeIncludingLocation:inStartingIndex withinRanges:inRangesArray];
	
	if (tRangeIndex!=NSNotFound)
	{
		NSValue * tValue;
		NSInteger tRangesCount;
		NSCharacterSet * tLookingForSet;
		NSString * tSourceCode;
		NSUInteger tBracesDepth=0;
		NSUInteger tParenthesisDepth=0;
		
		if (inCharacter=='(')
		{
			tParenthesisDepth=1;
		}
		else if (inCharacter=='{')
		{
			tBracesDepth=1;
		}
		else
		{
			return NSNotFound;
		}
		
		tRangeIndex++;
		
		tSourceCode=[textView_ string];
		
		tLookingForSet=[NSCharacterSet characterSetWithCharactersInString:@"(){}"];
		
		tRangesCount=[inRangesArray count];
		
		while (tRangeIndex>0)
		{
			NSRange tPartialLookRange;
			BOOL tContinue=YES;
			
			tValue=[inRangesArray objectAtIndex:(tRangeIndex-1)];
			
			tPartialLookRange=[tValue rangeValue];
			
			if (NSMaxRange(tPartialLookRange)>=inStartingIndex)
			{
				if (inStartingIndex==(tPartialLookRange.location))
				{
					tRangeIndex--;
					
					continue;
				}
				
				tPartialLookRange.length=inStartingIndex-tPartialLookRange.location;
			}
			
			while (tContinue==YES)
			{
				NSRange tFoundRange;
				
				tFoundRange=[tSourceCode rangeOfCharacterFromSet:tLookingForSet options:NSBackwardsSearch range:tPartialLookRange];
				
				if (tFoundRange.location==NSNotFound)
				{
					break;
				}
				else
				{
					unichar tFoundCharacter;
					
					tFoundCharacter=[tSourceCode characterAtIndex:tFoundRange.location];
					
					switch (tFoundCharacter)
					{
						case '}':
						
							if (tParenthesisDepth!=0)
							{
								return NSNotFound;
							}
						
							tBracesDepth++;
							
							break;
						
						case '{':
						
							if (tParenthesisDepth!=0)
							{
								return NSNotFound;
							}
						
							tBracesDepth--;
							
							if (tBracesDepth==0 && inCharacter=='{')
							{
								return tFoundRange.location;
							}
							
							break;
							
						case ')':
							
							tParenthesisDepth++;
							
							break;
							
						case '(':
							
							tParenthesisDepth--;
							
							if (tParenthesisDepth==0 && inCharacter=='(')
							{
								return tFoundRange.location;
							}
							
							break;
					}
					
					if (tFoundRange.location==tPartialLookRange.location)
					{
						tContinue=NO;
					}
					else
					{
						tPartialLookRange.length=tFoundRange.location-tPartialLookRange.location;
					}
				}
			}
			
			tRangeIndex--;
		}
	}
	
	return NSNotFound;
}

#pragma mark -

/*- (NSArray *) textView:(NSTextView *) inTextView completions:(NSArray *)words forPartialWordRange:(NSRange)charRange indexOfSelectedItem:(NSInteger *) inIndex
{
	// A COMPLETER
}*/

#pragma mark -

- (void) textViewDidChangeFont:(NSNotification *) inNotification
{
	if ([inNotification object]==textView_)
	{
		NSFont * tFont;
		NSUserDefaults * tUserDefaults;
		
		tUserDefaults=[NSUserDefaults standardUserDefaults];
		
		tFont=[textView_ font];
		
		[tUserDefaults setObject:[NSString stringWithFormat:@"%@ - %f",[tFont fontName],[tFont pointSize]]
						  forKey:IC_SOURCETEXTVIEW_DELEGATE_EDITOR_FONT];
	}
}

- (void) textDidChange:(NSNotification *) inNotification
{
    [NSObject cancelPreviousPerformRequestsWithTarget:self selector:@selector(delayedParseFunctions:) object:nil];
	
	[self parseSourceCode];
	
	[self performSelector:@selector(delayedParseFunctions:) withObject:nil afterDelay:0.5f];
}

- (void) textViewDidChangeSelection:(NSNotification *) inNotification
{
	if ([inNotification object]==textView_)
	{
		NSRange tSelectionRange;
		NSEvent * tCurrentEvent;
		
		tCurrentEvent=[NSApp currentEvent];
		
		tSelectionRange  = [textView_ selectedRange];
		
		if ([tCurrentEvent type] == NSLeftMouseUp && [tCurrentEvent clickCount] == 2 && tSelectionRange.length == 1)
		{
			NSString * tString;
			NSUInteger tLocation;
			NSUInteger tLength;
			
			tString=[textView_ string];
			
			tLength=[tString length];
			
			tLocation=tSelectionRange.location;
			
			if (tLocation>=0 && tLocation<tLength)
			{
				unichar tCharacter;
				unichar tOppositeCharacter=0;
				
				tCharacter=[tString characterAtIndex:tLocation];
			
				switch(tCharacter)
				{
					case '(':
					
						tOppositeCharacter=')';
						
						break;
					
					case '{':
						
						tOppositeCharacter='}';
						
						break;
						
					case '}':
					
						tOppositeCharacter='{';
						
						break;
					
					case ')':
						
						tOppositeCharacter='(';
						
						break;
					
				}
				
				if (tOppositeCharacter!=0)
				{
					if ([NSRangeUtilities location:tLocation isInsideRanges:searchableRangesArray_]==YES)
					{
						NSUInteger tMatchedLocation;
						
						if (tOppositeCharacter=='}' ||
							tOppositeCharacter==')')
						{
							tMatchedLocation=[self findForwardMatchingCharacter:tOppositeCharacter inRanges:searchableRangesArray_ startingAt:tLocation];
						}
						else
						{
							tMatchedLocation=[self findBackwardMatchingCharacter:tOppositeCharacter inRanges:searchableRangesArray_ startingAt:tLocation];
						}
						
						if (tMatchedLocation!=NSNotFound)
						{
							NSRange tSelectedRange;
							
							if (tOppositeCharacter=='}' ||
								tOppositeCharacter==')')
							{
								tSelectedRange=NSMakeRange(tLocation,tMatchedLocation-tLocation+1);
							}
							else
							{
								tSelectedRange=NSMakeRange(tMatchedLocation,tLocation-tMatchedLocation+1);
							}
							
							[textView_ setSelectedRange:tSelectedRange];
							
							[textView_ scrollRangeToVisible:tSelectedRange];
							
							return;
						}
					}
				
					NSBeep();
				}
			}
		}
		else
		{
			[NSObject cancelPreviousPerformRequestsWithTarget:self selector:@selector(updateFunctionsMenuSelectedItem:) object:nil];
			
			[self performSelector:@selector(updateFunctionsMenuSelectedItem:) withObject:nil afterDelay:IC_SOURCETEXTVIEW_DELEGATE_DELAYED_POPUP_UPDATE];
		}
	}
}

#pragma mark -

- (IBAction) showFunction:(id) sender
{
	NSString * tFunctionName;
	NSInteger tIndex;
	
	tIndex=[sender indexOfSelectedItem];
	
	tFunctionName=[sender titleOfSelectedItem];
	
	if (functions_!=nil && tIndex<[functions_ count])
	{
		NSDictionary * tFunctionDictionary;
		
		tFunctionDictionary=[functions_ objectAtIndex:tIndex];
		
		if ([[tFunctionDictionary objectForKey:FUNCTION_NAME_KEY] isEqualToString:tFunctionName]==NO)
		{
			NSEnumerator * tEnumerator;
		
			tEnumerator=[functions_ objectEnumerator];
			
			if (tEnumerator!=nil)
			{
				while (tFunctionDictionary=[tEnumerator nextObject])
				{
					NSString * tTitle;
					
					tTitle=[tFunctionDictionary objectForKey:FUNCTION_NAME_KEY];
					
					if ([tTitle isEqualToString:tFunctionName]==YES)
					{
						break;
					}
				}
			}
		}
		
		if (tFunctionDictionary!=nil)
		{
			NSValue * tValue;
			
			tValue=[tFunctionDictionary objectForKey:FUNCTION_PROTOTYPE_KEY];
			
			if (tValue!=nil)
			{
				NSRange tRange;
				NSRect tFrame;
				
				tRange=[tValue rangeValue];
				
				[textView_ setSelectedRange:tRange];
    
				[textView_ scrollRangeToVisible:tRange];
				
				tFrame=[functionsPopupButton_ frame];
				
				[functionsPopupButton_ sizeToFit];
				
				tFrame.size.width=NSWidth([functionsPopupButton_ frame])+5.0f;
				
				[functionsPopupButton_ setFrame:tFrame];
			
				[[functionsPopupButton_ superview] setNeedsDisplay:YES];
			}
		}
	}
}

- (NSArray *) parametersForFunctionNamed:(NSString *) inFunctionName
{
	if (functions_!=nil)
	{
		NSEnumerator * tEnumerator;
		
		tEnumerator=[functions_ objectEnumerator];
		
		if (tEnumerator!=nil)
		{
			NSDictionary * tFunctionDictionary;
			
			while (tFunctionDictionary=[tEnumerator nextObject])
			{
				NSString * tTitle;
				
				tTitle=[tFunctionDictionary objectForKey:FUNCTION_NAME_KEY];
				
				if ([tTitle isEqualToString:inFunctionName]==YES)
				{
					return [tFunctionDictionary objectForKey:FUNCTION_PARAMETERS_KEY];
				}
			}
		}
	}
	
	return nil;
}

- (NSArray *) functionsRanges
{
	NSMutableArray * tMutableArray;
	
	tMutableArray=[NSMutableArray array];
	
	if (tMutableArray!=nil)
	{
		NSEnumerator * tEnumerator;
		
		tEnumerator=[functions_ objectEnumerator];
		
		if (tEnumerator!=nil)
		{
			NSDictionary * tFunctionDictionary;
			
			while (tFunctionDictionary=[tEnumerator nextObject])
			{
				NSValue * tValue;
				
				tValue=[tFunctionDictionary objectForKey:FUNCTION_BLOCK_KEY];
				
				[tMutableArray addObject:tValue];
			}
		}
	}
	
	return tMutableArray;
}

- (NSArray *) sortedFunctionsList
{
	NSMutableArray * tMutableArray;
	
	tMutableArray=[NSMutableArray array];
	
	if (tMutableArray!=nil)
	{
		NSEnumerator * tEnumerator;
		
		tEnumerator=[functions_ objectEnumerator];
		
		if (tEnumerator!=nil)
		{
			NSDictionary * tFunctionDictionary;
			
			while (tFunctionDictionary=[tEnumerator nextObject])
			{
				NSString * tTitle;
				
				tTitle=[tFunctionDictionary objectForKey:FUNCTION_NAME_KEY];
				
				[tMutableArray addObject:tTitle];
			}
		}
		
		[tMutableArray sortUsingSelector:@selector(caseInsensitiveCompare:)];
	}
	
	return tMutableArray;
}

@end
